from .form_fields import *
from .types import *
from typing import get_origin, get_args, Union, Annotated
from .exceptions import *
from pydantic.fields import FieldInfo
import logging
from annotated_types import Gt, Lt, MinLen, MaxLen
import inspect, base64
from os import PathLike
from pathlib import Path
from datetime import datetime
import inspect
from fastapi import Form, UploadFile, Depends
from typing import Annotated, Sequence
logger = logging.getLogger(__name__)
logger.debug('Test message')


def unpack_annotation(annotation: type):
    while(is_union(annotation) or is_annotated(annotation)):
        annotation = get_args(annotation)[0]
    return annotation

def unpack_with_custom_annotation(annotation: type):
    while(is_custom(annotation) or is_union(annotation) or is_annotated(annotation)):
        annotation = get_args(annotation)[0]
    return annotation

def is_union(annotation: type):
    return get_origin(annotation) == Union

def is_list(annotation: type):
    return get_origin(annotation) == list or get_origin(annotation) == Sequence

def is_dict(annotation: type):
    return annotation == dict

def is_object(annotation: type):
    return inspect.isclass(annotation) and issubclass(annotation, FormModel)

def is_custom(annotation: type):
    return get_origin(annotation) == Custom

def is_datetime(annotation: type):
    return annotation == datetime 

def is_boolean(annotation: type):
    return annotation == FormBoolean or annotation == bool 

def get_object_type(annotation: type):
    args = get_args(annotation)
    if len(args):
        return args[0]
    return annotation

def get_list_item_type(annotation: type):
    return get_args(annotation)[0]

def is_number(annotation: type):
    return annotation == int or annotation == float or annotation == FormNumber 

def is_select(annotation: type):
    return inspect.isclass(annotation) and issubclass(annotation, Enum) 

def is_file(annotation: type):
    
    return inspect.isclass(annotation) and issubclass(annotation, Base64File) 

def is_literal(annotation: type):
    return get_origin(annotation) == Literal

def is_annotated(annotation: type):
    return get_origin(annotation) == Annotated
def is_text(annotation: type):
    return annotation == FormText

def get_validation_rules(field_name: str, field: FieldInfo):
    validation_rules = []
    for meta in field.metadata:
        if isinstance(meta, Gt):
            validation_rules.append(GreaterThan(value=meta.gt, error_text=f'{field_name} must be greater than {meta.gt}'))
        if isinstance(meta, Lt):
            validation_rules.append(LessThan(value=meta.lt, error_text=f'{field_name} must be less than {meta.lt}'))
        if isinstance(meta, MinLen):
            validation_rules.append(MinLength(length=meta.min_length, error_text=f'Minimum length of {field_name} is {meta.min_length}'))
        if isinstance(meta, MaxLen):
            validation_rules.append(MaxLength(length=meta.max_length, error_text=f'Maximum length of {field_name} is {meta.max_length}'))
    schema_data = field.json_schema_extra
    if schema_data is None:
        schema_data = {}
    if schema_data.get('required_if', None):
        validation_rules.append(RequiredIf(other_field_name=schema_data.get('required_if'), error_text=f'{field_name} is required'))
    if schema_data.get('required_unless', None):
        validation_rules.append(RequiredUnless(other_field_name=schema_data.get('required_unless'), error_text=f'{field_name} is required'))
    if schema_data.get('same_as', None):
        validation_rules.append(SameAs(other_field_name=schema_data.get('same_as'), error_text=f'{field_name} must be same as {schema_data.get("same_as")}'))
    if field.is_required():
        validation_rules.append(Required(error_text=f'{field_name} is required.'))
    return validation_rules

def to_form_field(field_name: str, field: FieldInfo)->FormField:
    annotation = field.annotation
    field_schema = field.json_schema_extra
    
    if field_schema is None:
        field_schema = {}
    validation_rules = get_validation_rules(field_schema.get('label', field_name), field)
    logger.info(f'schema: {field_schema}')
    field_definition = {
        'name': field_name,
        'validation_rules': validation_rules,
        'default': field.default,
        'meta': field.json_schema_extra
    } | field_schema
    logger.debug(f'{field_name} = annotation: {annotation}, schema: {field_definition}, validation rules: {validation_rules}')
    
    try:
        annotation = unpack_annotation(annotation)
        if is_custom(annotation):
            return CustomField.model_validate(field_definition)
        elif is_datetime(annotation):
            return DateTimeField.model_validate(field_definition)
        elif is_select(annotation):
            choices = []
            for enum_member in annotation:
                choices.append(enum_member.value)
            return SelectField.model_validate(field_definition | {'choices': choices})
        elif is_file(annotation):
            return FileField.model_validate(field_definition)
        elif is_list(annotation):
            list_item_type = get_list_item_type(annotation)
            field_info = field.from_annotated_attribute(list_item_type, default=None)
            field_definition['item_definition'] = to_form_field(field_name + '_item', field_info) 
            return ListField.model_validate(field_definition)
        elif is_object(annotation):
            field_definition['item_properties'] = get_object_type(annotation).get_form_fields() 
            return ObjectField.model_validate(field_definition)
        elif is_dict(annotation):
            logger.warning(f'dict is currently not supported')
            return None
        elif is_number(annotation):
            return NumberField.model_validate(field_definition)
        elif is_text(annotation):
            return TextField.model_validate(field_definition)
        elif is_literal(annotation):
            return TextField.model_validate(field_definition | {'rendered': False})
        elif is_boolean(annotation):
            return BooleanField.model_validate(field_definition)
        else:
            raise InvalidDefinitionException(f'Invalid field annotation {field_name}: {annotation}')
    except InvalidDefinitionException as e:
        e.message = f'Invalid field {field_name}: {e.message}'
        raise e

def to_multipart_form_field(field_name: str, field: FieldInfo):
    annotation = field.annotation
    try:
        annotation = unpack_with_custom_annotation(annotation)
        if is_object(annotation):
            pass
        elif is_list(annotation):
            pass
        elif is_file(annotation):
            pass
        
    except InvalidDefinitionException as e:
        e.message = f'Invalid field {field_name}: {e.message}'
        raise e

class FormModel(BaseSchema):
    @classmethod
    def get_form_fields(cls)->list[FormField]:
        fields = []
        for field_name, field_info in cls.model_fields.items():
            form_field = to_form_field(field_name, field_info)
            if form_field:
                fields.append(form_field)
        return fields
    
    @classmethod
    def as_multipart_form(cls):
        def __init__(self, **kwargs):
            # constructor for dynamically created classes.
            for k,v in kwargs.items():
                setattr(self, k, v)
        parameters = []
        annotations = {}
        for field_name, field_info in cls.model_fields.items():
            annotation = unpack_with_custom_annotation(field_info.annotation)
            field_annotation = Annotated[annotation, Form(...)]
            if is_object(annotation):
                sub_form = get_object_type(annotation).as_multipart_form()
                field_annotation = Annotated[sub_form, Depends()]
                # parameter_default = Depends(...)
            elif is_list(annotation):
                list_item_type = get_list_item_type(annotation)
                if is_object(list_item_type) or is_list(list_item_type):
                    raise InvalidDefinitionException(f'Field "{field_name}" in {cls.__name__}: Nested lists and lists of complex objects are not supported.')
                if is_file(list_item_type):
                    field_annotation = Annotated[list[bytes], list[UploadFile(...)]]
                else:
                    field_annotation = Annotated[list[list_item_type], list[Form(...)]]
                
            elif is_file(annotation):
                field_annotation = Annotated[bytes, UploadFile(...)]
            elif is_text(annotation):
                pass
            elif is_number(annotation):
                pass
            elif is_boolean(annotation):
                pass
            elif is_select(annotation):
                pass
            else:
                raise InvalidDefinitionException(f'Invalid field definition {field_name}: {annotation}')
            annotations[field_name] = field_annotation
            # inspect.Parameter(inspect.Parameter.)
            parameters.append(
                inspect.Parameter(
                    name=field_name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=inspect.Parameter.empty,
                    annotation=field_annotation
                )
            )
        class_name = f'Multipart{cls.__name__}'
        create_parameters = {
            '__init__': __init__,
            '__annotations__': annotations,
            '__signature__': inspect.Signature(parameters)
        }
        print(f'Signature: {inspect.Signature(parameters)}')
        print(f'Create {class_name} as {create_parameters}')
        return type(class_name,(object,),create_parameters )

    def save_file(self, directory: PathLike, file: Base64File):
        file_data = file
        file_data = Base64FileData.model_validate(file_data)
        if file_data.data:
            file_path = f'{directory}/{file_data.name}'
            with open(file_path, 'wb') as f:
                f.write(base64.b64decode(file_data.data))
            file.path = file_path

    def file_data_fields(self):
        file_data_fields = []
        for field_name, field_info in self.model_fields.items():
            annotation = field_info.annotation
            annotation = unpack_with_custom_annotation(annotation)
            
            if is_list(annotation):
                list_item_type = get_list_item_type(annotation)
                if is_union(list_item_type):
                    list_item_type = unpack_annotation(list_item_type)
                logger.debug(f'list with child item type: {list_item_type}')
                if is_file(list_item_type):

                    for file_data in getattr(self, field_name):
                        file_data_fields.append(file_data)
                elif is_object(list_item_type):
                    for item in getattr(self, field_name):
                        file_data_fields += item.file_data_fields()
            elif is_file(annotation):
                file_data: Base64File = getattr(self, field_name)
                if file_data:
                    file_data_fields.append(file_data)
            elif is_object(annotation):
                if getattr(self, field_name):
                    file_data_fields += getattr(self, field_name).file_data_fields()
        return file_data_fields

    def remove_file_data(self):
        for file_data_field in self.file_data_fields():
            file_data_field.data = None
        return self
    
    def load_files(self, allow_not_stored: bool = True):
        for file_field in self.file_data_fields():
            if not file_field.path:
                if allow_not_stored:
                    logger.warning(f'{file_field.name} has no stored path. skipping')
                    continue
                raise Exception(f'{file_field.name} has no path and is not stored on disk. Consider setting allow_not_stored = True or make sure that path is set.')
            with open(Path(file_field.path), 'rb') as f:
                file_field.data = base64.b64encode(f.read()).decode()
        return self
    
    def load_file_data(self, directory: PathLike):
        for file_data_field in self.file_data_fields():
            file_path = Path(directory).joinpath(file_data_field.name)
            with open(file_path, 'rb') as f:
                file_data_field.data = base64.b64encode(f.read()).decode()
            file_data_field.path = file_path
        return self

    def save_files(self, directory: PathLike):
        for file_data_field in self.file_data_fields():
            self.save_file(directory, file_data_field)
        return self
    
    def delete_files_from_directory(self, directory: PathLike, missing_ok: bool = False):
        for file_field in self.file_data_fields():
            file_path = Path(directory).joinpath(file_field.name)
            file_path.unlink(missing_ok=missing_ok)
        return self
    
    def delete_files(self, missing_ok: bool = False):
        for file_field in self.file_data_fields():
            file_path = Path(file_field.path)
            file_path.unlink(missing_ok=missing_ok)
        return self



import types
import typing
import annotated_types
import typing_extensions
from pydantic import AliasChoices, AliasPath, BaseModel, ConfigDict, TypeAdapter, Field, types
from pydantic_core import PydanticUndefined
from pydantic.config import JsonDict
from pydantic.fields import _Unset, _EmptyKwargs, FieldInfo, Deprecated
from pydantic_core.core_schema import IntSchema, FloatSchema, StringSchema, ModelSchema, ListSchema, CoreSchemaType
from typing import Optional, TypeVar, Literal, Type, IO, Annotated, Unpack
from enum import Enum
import base64
from io import BytesIO
from .types import *
from humps import camelize
from typing import Union, ForwardRef, Any

PydanticModelType = TypeVar('PydanticModelType', bound=BaseModel)

T = TypeVar('T')

DataSource = str

class BaseSchema(BaseModel):
    model_config: ConfigDict = ConfigDict(alias_generator=camelize, populate_by_name=True)



# FormValue = FormList|FormObject|FormText|FormNumber|FormSelect|FormBoolean|FormFile
class ValidationRuleName(str, Enum):
    REQUIRED = 'required'
    REQUIRED_IF = 'required_if'
    REQUIRED_UNLESS = 'required_unless'
    SAME_AS = 'same_as' 
    MIN_LENGTH = 'min_length'
    MAX_LENGTH = 'max_length'
    GREATER_THAN = 'greater_than'
    LESS_THAN = 'less_than'

class ValidationRule(BaseSchema):
    name: ValidationRuleName
    error_text: str
    other_field_name: Optional[str] = None
    length: Optional[int] = None
    value: Optional[float] = None
    has_value: Optional[bool] = None
    
class Required(ValidationRule):
    name: Literal[ValidationRuleName.REQUIRED] = ValidationRuleName.REQUIRED

class RequiredIf(Required):
    name: Literal[ValidationRuleName.REQUIRED_IF] = ValidationRuleName.REQUIRED_IF
    other_field_name: str
    
class RequiredUnless(RequiredIf):
    name: Literal[ValidationRuleName.REQUIRED_UNLESS] = ValidationRuleName.REQUIRED_UNLESS

class SameAs(ValidationRule):
    name: Literal[ValidationRuleName.SAME_AS] = ValidationRuleName.SAME_AS

class MinLength(ValidationRule):
    name: Literal[ValidationRuleName.MIN_LENGTH] = ValidationRuleName.MIN_LENGTH
    length: int

class MaxLength(MinLength):
    name: Literal[ValidationRuleName.MAX_LENGTH] = ValidationRuleName.MAX_LENGTH

class GreaterThan(ValidationRule):
    name: Literal[ValidationRuleName.GREATER_THAN] = ValidationRuleName.GREATER_THAN
    value: float

class LessThan(GreaterThan):
    name: Literal[ValidationRuleName.LESS_THAN] = ValidationRuleName.LESS_THAN

class RenderCondition(BaseSchema):
    property_path: str
    has_value: Any = None
    render_conditions: list['RenderCondition'] = []

class FormFieldType(str, Enum):
    OBJECT = 'object'
    LIST = 'list'
    NUMBER = 'number'
    TEXT = 'text'
    DATETIME = 'datetime'
    SELECT = 'select'
    FILE = 'file'
    BOOLEAN = 'boolean'


class FormField(BaseSchema):
    model_config: ConfigDict = ConfigDict(alias_generator=lambda name: camelize(name), populate_by_name=True)
    field_type: FormFieldType|str
    name: str
    hint: Optional[str] = None
    style: Optional[str] = None
    field_index: Optional[int] = int(1e7)
    default: Optional[Any] = None
    rendered: Optional[bool] = True
    render_conditions: list[RenderCondition] = []
    label: Optional[str] = None
    validation_rules: list[ValidationRule] = []
    item_definition: Optional[Any] = None
    choices: Optional[list[Any]] = None
    data_source: Optional[DataSource] = None
    item_value: Optional[str] = None
    radios: Optional[bool] = None
    item_text: Optional[str] = None
    item_properties: Optional[list[Any]] = None
    meta: Optional[dict[str, Any]] = {}

class NumberField(FormField):
    field_type: Literal[FormFieldType.NUMBER] = FormFieldType.NUMBER

class TextField(FormField):
    field_type: Literal[FormFieldType.TEXT] = FormFieldType.TEXT

class DateTimeField(FormField):
    field_type: Literal[FormFieldType.DATETIME] = FormFieldType.DATETIME 


class FileField(FormField):
    field_type: Literal[FormFieldType.FILE] = FormFieldType.FILE

class SelectField(FormField):
    field_type: Literal[FormFieldType.SELECT] = FormFieldType.SELECT
    choices: Optional[list[Any]] = None
    data_source: Optional[DataSource] = None
    item_value: Optional[str] = None
    item_text: Optional[str] = None
    radios: Optional[bool] = None

class BooleanField(FormField):
    field_type: Literal[FormFieldType.BOOLEAN] = FormFieldType.BOOLEAN
    tri_state: Optional[bool] = False

ListField = ForwardRef('ListField')
class CustomField(FormField):
    field_type: str

class ObjectField(FormField):
    field_type: Literal[FormFieldType.OBJECT] = FormFieldType.OBJECT
    item_properties: list[Any]

class ListField(FormField):
    field_type: Literal[FormFieldType.LIST] = FormFieldType.LIST
    item_definition: Any

ListField.model_rebuild()


def FField(  # noqa: C901
    default: Any = PydanticUndefined,
    default_factory: typing.Callable[[], Any] | None = _Unset,
    alias: str | None = _Unset,
    field_type: FormFieldType | None = _Unset,
    hint: Optional[str] = _Unset,
    style: Optional[str] = _Unset,
    field_index: Optional[int] = int(1e7),
    rendered: Optional[bool] = True,
    render_conditions: list[RenderCondition] = [],
    label: Optional[str] = _Unset,
    validation_rules: Optional[list[ValidationRule]] = [],
    choices: Optional[list[Any]] = _Unset,
    radios: Optional[bool] = _Unset,
    alias_priority: int | None = _Unset,
    validation_alias: str | AliasPath | AliasChoices | None = _Unset,
    serialization_alias: str | None = _Unset,
    title: str | None = _Unset,
    field_title_generator: typing_extensions.Callable[[str, FieldInfo], str] | None = _Unset,
    description: str | None = _Unset,
    examples: list[Any] | None = _Unset,
    exclude: bool | None = _Unset,
    discriminator: str | types.Discriminator | None = _Unset,
    deprecated: Deprecated | str | bool | None = _Unset,
    json_schema_extra: JsonDict | typing.Callable[[JsonDict], None] | None = _Unset,
    frozen: bool | None = _Unset,
    validate_default: bool | None = _Unset,
    repr: bool = _Unset,
    init: bool | None = _Unset,
    init_var: bool | None = _Unset,
    kw_only: bool | None = _Unset,
    pattern: str | typing.Pattern[str] | None = _Unset,
    strict: bool | None = _Unset,
    coerce_numbers_to_str: bool | None = _Unset,
    gt: annotated_types.SupportsGt | None = _Unset,
    ge: annotated_types.SupportsGe | None = _Unset,
    lt: annotated_types.SupportsLt | None = _Unset,
    le: annotated_types.SupportsLe | None = _Unset,
    multiple_of: float | None = _Unset,
    allow_inf_nan: bool | None = _Unset,
    max_digits: int | None = _Unset,
    decimal_places: int | None = _Unset,
    min_length: int | None = _Unset,
    max_length: int | None = _Unset,
    union_mode: Literal['smart', 'left_to_right'] = _Unset,
    fail_fast: bool | None = _Unset,
    *args,
    **kwargs
) -> Any:
    """Usage docs: https://docs.pydantic.dev/2.9/concepts/fields
    (Extended with parameters for pydantic_form_model fields)
    Create a field for objects that can be configured.

    Used to provide extra information about a field, either for the model schema or complex validation. Some arguments
    apply only to number fields (`int`, `float`, `Decimal`) and some apply only to `str`.

    Note:
        - Any `_Unset` objects will be replaced by the corresponding value defined in the `_DefaultValues` dictionary. If a key for the `_Unset` object is not found in the `_DefaultValues` dictionary, it will default to `None`

    Args:
        default: Default value if the field is not set.
        default_factory: A callable to generate the default value, such as :func:`~datetime.utcnow`.
        alias: The name to use for the attribute when validating or serializing by alias.
            This is often used for things like converting between snake and camel case.
        field_type: The field type for this field (use this to overwrite the inferred field type from the type annotation),
        hint: Hint for the field (usually a short description),
        style: CSS stylings to store for this field,
        field_index: Index property (can be used to sort fields),
        rendered: Wether this field is rendered by default or not,
        render_conditions: List of render conditions,
        label: Text label for this fields,
        validation_rules: A list of additional validation rules for this field,
        choices: Choices for this field (can be used to store possible selections),
        alias_priority: Priority of the alias. This affects whether an alias generator is used.
        validation_alias: Like `alias`, but only affects validation, not serialization.
        serialization_alias: Like `alias`, but only affects serialization, not validation.
        title: Human-readable title.
        field_title_generator: A callable that takes a field name and returns title for it.
        description: Human-readable description.
        examples: Example values for this field.
        exclude: Whether to exclude the field from the model serialization.
        discriminator: Field name or Discriminator for discriminating the type in a tagged union.
        deprecated: A deprecation message, an instance of `warnings.deprecated` or the `typing_extensions.deprecated` backport,
            or a boolean. If `True`, a default deprecation message will be emitted when accessing the field.
        json_schema_extra: A dict or callable to provide extra JSON schema properties.
        frozen: Whether the field is frozen. If true, attempts to change the value on an instance will raise an error.
        validate_default: If `True`, apply validation to the default value every time you create an instance.
            Otherwise, for performance reasons, the default value of the field is trusted and not validated.
        repr: A boolean indicating whether to include the field in the `__repr__` output.
        init: Whether the field should be included in the constructor of the dataclass.
            (Only applies to dataclasses.)
        init_var: Whether the field should _only_ be included in the constructor of the dataclass.
            (Only applies to dataclasses.)
        kw_only: Whether the field should be a keyword-only argument in the constructor of the dataclass.
            (Only applies to dataclasses.)
        coerce_numbers_to_str: Whether to enable coercion of any `Number` type to `str` (not applicable in `strict` mode).
        strict: If `True`, strict validation is applied to the field.
            See [Strict Mode](../concepts/strict_mode.md) for details.
        gt: Greater than. If set, value must be greater than this. Only applicable to numbers.
        ge: Greater than or equal. If set, value must be greater than or equal to this. Only applicable to numbers.
        lt: Less than. If set, value must be less than this. Only applicable to numbers.
        le: Less than or equal. If set, value must be less than or equal to this. Only applicable to numbers.
        multiple_of: Value must be a multiple of this. Only applicable to numbers.
        min_length: Minimum length for iterables.
        max_length: Maximum length for iterables.
        pattern: Pattern for strings (a regular expression).
        allow_inf_nan: Allow `inf`, `-inf`, `nan`. Only applicable to numbers.
        max_digits: Maximum number of allow digits for strings.
        decimal_places: Maximum number of decimal places allowed for numbers.
        union_mode: The strategy to apply when validating a union. Can be `smart` (the default), or `left_to_right`.
            See [Union Mode](../concepts/unions.md#union-modes) for details.
        fail_fast: If `True`, validation will stop on the first error. If `False`, all validation errors will be collected.
            This option can be applied only to iterable types (list, tuple, set, and frozenset).

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo]. The return annotation is `Any` so `Field` can be used on
            type-annotated fields without causing a type error.
    """
    return Field(
        default = default,
        default_factory = default_factory,
        alias = alias,
        field_type = field_type,
        hint = hint,
        style = style,
        field_index = field_index,
        rendered = rendered,
        render_conditions = render_conditions,
        label = label,
        validation_rules = validation_rules,
        choices = choices,
        radios=radios,
        alias_priority = alias_priority,
        validation_alias = validation_alias,
        serialization_alias = serialization_alias,
        title = title,
        field_title_generator = field_title_generator,
        description = description,
        examples = examples,
        exclude = exclude,
        discriminator = discriminator,
        deprecated = deprecated,
        json_schema_extra = json_schema_extra,
        frozen = frozen,
        validate_default = validate_default,
        repr = repr,
        init = init,
        init_var = init_var,
        kw_only = kw_only,
        pattern = pattern,
        strict = strict,
        coerce_numbers_to_str = coerce_numbers_to_str,
        gt = gt,
        ge = ge,
        lt = lt,
        le = le,
        multiple_of = multiple_of,
        allow_inf_nan = allow_inf_nan,
        max_digits = max_digits,
        decimal_places = decimal_places,
        min_length = min_length,
        max_length = max_length,
        union_mode = union_mode,
        fail_fast = fail_fast,
        *args, **kwargs)


from typing import Any, TypeVar, Generic

from pydantic_core import core_schema
from typing_extensions import get_args
from pydantic import BaseModel, GetCoreSchemaHandler
from typing import get_origin, Optional
from os import PathLike

T = TypeVar('T')


class File(Generic[T]):

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        origin = get_origin(source_type)
        if origin is None:  # used as `x: Owner` without params
            origin = source_type
            item_tp = Any
        else:
            item_tp = get_args(source_type)[0]
        # both calling handler(...) and handler.generate_schema(...)
        # would work, but prefer the latter for conceptual and consistency reasons
        item_schema = handler.generate_schema(item_tp)
        return item_schema
    
    def save(self, path: PathLike):
        raise Exception('Not implemented')

class Base64FileData(BaseModel):
    data: Optional[str] = None
    name: Optional[str] = None
    

class Base64File(Base64FileData, File[Base64FileData]):
    path: Optional[str|PathLike] = None

class Custom(Generic[T]):

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        origin = get_origin(source_type)
        if origin is None:  # used as `x: Owner` without params
            origin = source_type
            item_tp = Any
        else:
            item_tp = get_args(source_type)[0]
        # both calling handler(...) and handler.generate_schema(...)
        # would work, but prefer the latter for conceptual and consistency reasons
        item_schema = handler.generate_schema(item_tp)
        return item_schema
    

FormNumber = int|float
FormText = str
FormBoolean = bool



class FormModelException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

class InvalidDefinitionException(FormModelException):
    pass